/* Linux console support for xz80, copyright (C) 1994/5 Russell Marks.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef LINUX_SVGALIB

/* an explanation of the way graphics works:
 * we bang the border up around the speccy screen when it's changed,
 * but otherwise just draw the whole 256x192 on each reqd. refresh.
 *
 * I tried my original idea (below) first, but 'twas WWAAAAYYYY too slow.
 *
 * We use image to store a malloc'ed 256x192 XImage-like image of
 * the speccy screen. Whenever a screen refresh is done, we blank out the
 * a virtual screen svgalib GC in the border colour, then put the speccy
 * screen image into that, *then* copy that virtual screen to the real one.
 * It has to be done this way, as the functions available for planar
 * 256-colour modes are very limited.
 *
 * Er, but it's done that way again now, except it's unbuffered.
 */
 


#include <stdio.h>
#include <signal.h>
#include <stdlib.h>	/* or could use malloc.h */
#include <string.h>
#include <unistd.h>
#include <termios.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/kd.h>
#include <linux/keyboard.h>	/* mainly for NR_KEYS */
#include <linux/vt.h>
#include <vga.h>
#include <vgagl.h>
#include "rawkey.h"
/*#include "resource.h"*/
#include "linuxcon.h"
#include "z80.h"


static int keymap[NR_KEYS];	/* scancode -> ASCII translation */
static int asciimap[256];	/* ASCII -> scancode translation */
static int tty_fd=-1;

static int old_border = -1;
static int shift_pressed = 0,ctrl_pressed = 0,alt_pressed=0;
static int colour_index_kludge = 0;
static int palette_sav[16*3];
static int restart_emu;

extern unsigned char *image;
extern int bordercr,refresh_screen;
extern unsigned char keyports[];

static struct termios new_termio,old_termio;

XrmOptionDescRec *option_table;
int option_size;

char resource_stuff[1024];

#define MAX_OPTS	100

/* we store the chosen options in here, as 'strings' in numeric order as
 * dictated by opts[].
 */
char stored_options[MAX_OPTS][80];


void
linux_get_keys(XEvent *xev)
{
int c;

/* we use BFI to fix the PrtSc/Pause problem - i.e. we don't :^) */
while((c=get_scancode())==0xE0);
if(c==0xE1) c=get_scancode();
	
if(c!=-1)
  {
  xev->type = (c&128)?KeyRelease:KeyPress;
		c &= 127;
		xev->key = c;
		
		if (c == LEFT_SHIFT || c == RIGHT_SHIFT)
			/* this isn't quite right, but it's good enough */
			shift_pressed = (xev->type == KeyPress)?1:0;

		if (c == LEFT_CTRL)
			ctrl_pressed  = (xev->type == KeyPress)?1:0;
		
		if (c == LEFT_ALT)
			alt_pressed  = (xev->type == KeyPress)?1:0;
		
		/* these are for a UK keyboard. It's possible to do this
		 * better :), but it's pretty awkward.
		 */
		if (shift_pressed) {
			/* with shift */
			if (c == scancode_trans('-')) xev->key = XK_underscore;
			if (c == scancode_trans(';')) xev->key = XK_colon;
			if (c == scancode_trans(',')) xev->key = XK_less;
			if (c == scancode_trans('.')) xev->key = XK_greater;
			if (c == scancode_trans('/')) xev->key = XK_question;
			if (c == scancode_trans('\\')) xev->key = XK_bar;
		} else {
			/* without shift */
			if (c == scancode_trans('=')) xev->key = XK_equal;
			if (c == scancode_trans(';')) xev->key = XK_semicolon;
			if (c == scancode_trans('\'')) xev->key = XK_apostrophe;
			if (c == scancode_trans(',')) xev->key = XK_comma;
			if (c == scancode_trans('.')) xev->key = XK_period;
			if (c == scancode_trans('/')) xev->key = XK_slash;
			if (c == scancode_trans('\\')) xev->key = XK_backslash;
		}
#ifdef CONTROL_Z_SUSPENDS
		if (ctrl_pressed && c == scancode_trans('z'))
			xev->key = CTRL_Z;
#endif
		if (alt_pressed && c>=XK_F1 && c<=XK_F9+1) {
			struct vt_stat vts;
			int newvt;
			
			ioctl(tty_fd,VT_GETSTATE,&vts);
			newvt=c-XK_F1+1;
			if(vts.v_active!=newvt) {
				ioctl(tty_fd,VT_ACTIVATE,newvt);
				restart_emu=0;
				while(restart_emu==0) usleep(50000);
			}
			xev->type=KeyRelease;
			xev->key=NoKeyPressed;
		}
	} else {
		xev->type = KeyRelease;
		xev->key  = NoKeyPressed;
	}
}


void
linux_panic_now(void)
{
linux_graphics_off();
fprintf(stderr,"xz80: caught a segfault!\n");
exit(1);
}


void
linux_graphics_on(void)
{
	vga_disabledriverreport();
	vga_init();
	vga_setmode(MODE_TO_USE);
	signal(SIGSEGV,(void (*)(int))linux_panic_now);
	gl_setcontextvga(MODE_TO_USE);

	rawmode_init();

	/* fix termio stuff so ^C-style interrupts are ignored */
	ioctl(tty_fd,TCGETS,&old_termio);
	new_termio=old_termio;
	new_termio.c_lflag&=~(ISIG|ICANON);
	ioctl(tty_fd,TCSETSW,&new_termio);
	
	if ((image = malloc(256*192)) == NULL) {
		linux_graphics_off();
		fprintf(stderr,"xz80: not enough memory for image!\n");
		exit(1);
	}
}


void
linux_graphics_off(void)
{
rawmode_exit();
ioctl(tty_fd,TCSETSW,&old_termio);
vga_setmode(TEXT);
if (image != NULL) free(image);
}


void linux_draw_border()
{
gl_fillbox(0,0,MODE_XSIZE,SPECSCRN_YPOS,bordercr);
gl_fillbox(0,SPECSCRN_YPOS,SPECSCRN_XPOS,192,bordercr);
gl_fillbox(SPECSCRN_XPOS+256,SPECSCRN_YPOS,
  MODE_XSIZE-(SPECSCRN_XPOS+256),192,bordercr);
gl_fillbox(0,SPECSCRN_YPOS+192,MODE_XSIZE,
  MODE_YSIZE-(SPECSCRN_YPOS+192),bordercr);
}


void XShmPutImage(z1,z2,z3,ptr,z4,z5,x,y,w,h,z6)
int z1,z2,z3,z4,z5,z6;
unsigned char *ptr;
int x,y,w,h;
{
gl_putboxpart(SPECSCRN_XPOS+x,SPECSCRN_YPOS+y,w,h,256,192,image,x,y);
}

void XPutImage(z1,z2,z3,ptr,z4,z5,x,y,w,h)
int z1,z2,z3,z4,z5;
unsigned char *ptr;
int x,y,w,h;
{
gl_putboxpart(SPECSCRN_XPOS+x,SPECSCRN_YPOS+y,w,h,256,192,image,x,y);
}


int XAllocColor(Display *dpy,Colormap cmap,XColor *xc)
{
vga_setpalette(xc->pixel=colour_index_kludge++,
	(xc->red)>>10,(xc->green)>>10,(xc->blue)>>10);
return(1);
}


void raw_mode(tty_fd,on)
int tty_fd,on;
{
ioctl(tty_fd,KDSKBMODE,on?K_RAW:K_XLATE);
}


void
linux_temp_text(void)
{
ioctl(tty_fd,TCSETSW,&old_termio);
rawmode_exit();
vga_getpalvec(0,16,palette_sav);
vga_setmode(TEXT);
}


void
linux_text_end(void)
{
#if 0
	usleep(500000);		/* let them read any message */
#endif
	vga_setmode(MODE_TO_USE);
	vga_setpalvec(0,16,palette_sav);
	old_border=-1;
        
        refresh_screen=1;
	
	shift_pressed=ctrl_pressed=alt_pressed=0;
	keyports[0] |= 0x01;
	keyports[7] |= 0x02;

	rawmode_init();
	ioctl(tty_fd,TCSETSW,&new_termio);
}


void vt_from_speccy()
{
ioctl(tty_fd,TCSETSW,&old_termio);
vga_getpalvec(0,16,palette_sav);
vga_setmode(TEXT);
raw_mode(tty_fd,0);  /* don't use rawmode_exit 'cos of other things it does */
ioctl(tty_fd,VT_RELDISP,VT_ACKACQ);
signal(SIGUSR1,vt_from_speccy);
}


void vt_to_speccy()
{
linux_text_end();
restart_emu=1;		/* we're back, say to start up again */
signal(SIGUSR2,vt_to_speccy);
}


/* it's really easy to translate the scancodes these days, we just
 * use the keytable stuff!
 *
 * returns 0 on error, 1 if ok
 */
int get_keyb_map()
{
struct kbentry keyb_ent;
int f;

keyb_ent.kb_table=0;	/* unshifted */
for(f=0;f<NR_KEYS;f++)
  {
  keyb_ent.kb_index=f;
  
  if(ioctl(tty_fd,KDGKBENT,(unsigned int)&keyb_ent))
    return(0);
    
  keymap[f]=keyb_ent.kb_value;
  }
return(1);
}


void make_ascii_map()
{
int f,i,found;

for(f=0;f<256;f++)
  {
  found=0;
  for(i=0;i<NR_KEYS;i++)
    if((KTYP(keymap[i])==KT_LATIN || KTYP(keymap[i])==KT_LETTER ||
        KTYP(keymap[i])==KT_ASCII) && KVAL(keymap[i])==f)
      {
      found=1;
      break;
      }
      
  asciimap[f]=found?i:-1;
  }
}


/* returns 1 if ok, 0 otherwise */
int rawmode_init()
{
if(tty_fd==-1)
  if((tty_fd=open("/dev/tty",O_RDONLY | O_NONBLOCK))<0)
    return(0);

if(get_keyb_map(tty_fd,keymap))
  {
  struct vt_mode vtm;
  
  make_ascii_map();
  raw_mode(tty_fd,1);
  signal(SIGUSR1,vt_from_speccy);
  signal(SIGUSR2,vt_to_speccy);
  ioctl(tty_fd,VT_GETMODE,&vtm);
  vtm.mode=VT_PROCESS;
  vtm.relsig=SIGUSR1;
  vtm.acqsig=SIGUSR2;
  ioctl(tty_fd,VT_SETMODE,&vtm);
  return(1);
  }
else
  return(0);
}


void rawmode_exit()
{
struct vt_mode vtm;

ioctl(tty_fd,VT_GETMODE,&vtm);
vtm.mode=VT_AUTO;
ioctl(tty_fd,VT_SETMODE,&vtm);
raw_mode(tty_fd,0);
close(tty_fd);
tty_fd=-1;
}


/* returns -1 if no keypresses pending, else returns scancode. */
int get_scancode()
{
unsigned char c;

if(read(tty_fd,&c,1)<=0)
  return(-1);

return((int)c);
}


/* converts scancode to key binding */
int keymap_trans(sc)
int sc;
{
if(sc<0 || sc>127) return(-1);
return(keymap[sc]);
}


/* converts ASCII to scancode */
int scancode_trans(asc)
int asc;
{
if(asc<0 || asc>255) return(-1);
return(asciimap[asc]);
}


void
XrmInitialize(void)
{
int f;

for(f=0;f<MAX_OPTS;f++)
  stored_options[f][0]=0;
}


void
delete_arg(int *argcp,char **argv)
{
int f;

/* remember that at the moment, *argcp is one less than an argc would
 * normally be. Also, it's part of the way through the args.
 */
for(f=0;f<*argcp-1;f++)
  argv[f]=argv[f+1];

argv[*argcp-1]=NULL;
(*argcp)--;
}


void
XrmParseCommand(XrmDatabase *d,XrmOptionDescRec *o,int onum,char *name,
						int *argcp,char **argv)
{
XrmOptionDescRec *recptr;
int f;
int grokking,found;
int argc;

option_table=o;
option_size=onum;

grokking=1;

argv++; (*argcp)--;
while(grokking && *argcp>0)
  {
  recptr=option_table;
  for(f=0,found=0;f<option_size && !found;f++,recptr++)
    {
    if(strcasecmp(*argv,recptr->option)==0)
      {
      delete_arg(argcp,argv);
      found=1;
      if(recptr->argKind==XrmoptionNoArg)
        strcpy(stored_options[f],recptr->option+1);
      else
        {
        if(**argv=='-' || (*argcp)==0)
          found=grokking=0;
        else
          {
          strcpy(stored_options[f],*argv);
          delete_arg(argcp,argv);
          }
        }
      }
    }
  
  if(found==0) grokking=0;
  }

(*argcp)++;
}


/* the option name is in 'class' */
int
XrmGetResource(XrmDatabase d,char *name,char *class,char **type,XrmValue *val)
{
XrmOptionDescRec *recptr;
int f;

val->addr=NULL;
for(f=0,recptr=option_table;f<option_size;f++,recptr++)
  {
  /* XXX this is a fairly unpleasant kludge */
  if(strcmp(recptr->option+1,class+5)==0 && stored_options[f][0]!=0)
    {
    val->addr=stored_options[f];
    val->size=strlen(val->addr);
    return(1);
    }
  }
return(0);
}

#endif	/* LINUX_SVGALIB */
